#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"

typedef unsigned char u8;

#define MIN_RUN     3
#define MAX_RUN     (127 + MIN_RUN)
#define MIN_STRING  1
#define MAX_STRING  (127 + MIN_STRING)

void compress(u8* input, int& inputSize, u8* output, int& outputSize)
{
    int stringStart = 0;
    int runStart = 0;
    int currentByte = 1;
    outputSize = 0;
    bool isInRun = false;

    while (true)
    {
        if (isInRun == false)
        {
            // string mode
            if (input[currentByte] != input[runStart])
                runStart = currentByte;

            int stringLength = runStart - stringStart;
            int runLength    = currentByte + 1 - runStart;

            if (currentByte == inputSize ||
                runLength    == MIN_RUN ||
                stringLength == MAX_STRING)
            {
                // emit string
                if (stringLength >= MIN_STRING)
                {
                    output[outputSize++] = stringLength - MIN_STRING;

                    for (int i = 0; i < stringLength; i++)
                        output[outputSize++] = input[stringStart++];
                }

                // go to run mode
                if (runLength == MIN_RUN)
                    isInRun = true;
            }
        }
        else
        {
            // run mode
            stringStart = currentByte;

            int runLength = currentByte - runStart;

            if (input[currentByte] != input[runStart] ||
                currentByte == inputSize ||
                runLength == MAX_RUN)
            {
                // emit run
                output[outputSize++] = 128 + runLength - MIN_RUN;
                output[outputSize++] = input[runStart];
                runStart = currentByte;

                // go to string mode
                isInRun = false;
            }
        }

        if (currentByte == inputSize)
            break;

        currentByte++;
    }
}

void uncompress(u8* input, int& inputSize, u8* output, int& outputSize)
{
    int i = 0;
    outputSize = 0;

    while (i != inputSize)
    {
        if (input[i] < 128)
        {
            int length = input[i++] + MIN_STRING;

            for (int j = 0; j < length; j++)
                output[outputSize++] = input[i++];
        }
        else
        {
            int length = input[i++] - 128 + MIN_RUN;

            for (int j = 0; j < length; j++)
                output[outputSize++] = input[i];

            i++;
        }
    }
}

int fileGetSize(char* fileName)
{
    int fileSize;

    FILE* f = fopen(fileName, "rb");
    fseek(f, 0, SEEK_END);
    fileSize = ftell(f);
    fclose(f);
    return fileSize;
}

int fileLoad(char* fileName, u8** mem)
{
    int fileSize = fileGetSize(fileName);

    *mem = (u8*)malloc(fileSize);
    FILE* f = fopen(fileName, "rb");
    fread(*mem, 1, fileSize, f);
    fclose(f);
    return fileSize;
}

void fileSave(char* fileName, u8* mem, int fileSize, int originalSize)
{
    FILE* f = fopen(fileName, "wb");

    if (originalSize != 0)
        fwrite(&originalSize, sizeof(int), 1, f);

    fwrite(mem, 1, fileSize, f);
    fclose(f);
}

void fileCompress(char* srcName, char* destName)
{
    u8* input;
    int inputSize;
    u8* output;
    int outputSize = 0;

    inputSize = fileLoad(srcName, &input);
    printf("original size: %d\n", inputSize);

    output = (u8*)malloc(inputSize * 2);
    compress(input, inputSize, output, outputSize);

    fileSave(destName, output, outputSize, inputSize);
    printf("compressed size: %d\n", outputSize);

    free(input);
    free(output);
}

void fileUncompress(char* srcName, char* destName)
{
    u8* input;
    int inputSize;
    u8* output;
    int outputSize = 0;

    inputSize = fileLoad(srcName, &input);

    outputSize = *((int*)input);
    output = (u8*)malloc(outputSize);
    inputSize -= sizeof(int);
    uncompress(input + sizeof(int), inputSize, output, outputSize);

    fileSave(destName, output, outputSize, 0);
    printf("uncompressed size: %d\n", outputSize);

    free(input);
    free(output);
}

int main(int argc, char* argv[])
{
    fileCompress("image.tga", "comp.rle");
    fileUncompress("comp.rle", "uncomp.tga");

    return 0;
}
